# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Gettings Things Gnome! - a personal organizer for the GNOME desktop
# Copyright (c) 2008-2009 - Lionel Dricot & Bertrand Rousseau
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------


"""
The core functionality GTG.

In order to not interact directly with the datastore, we provide
"requesters".  The requester is only an interface and there can be as
many requester as you want as long as they are all from the same
datastore.  Requester also provides an interface for the tagstore

If you want to display only a subset of tasks, you can either:

 - have access to the main FilteredTree (the one displayed in the main
   window) and apply filters on it.  (You can create your own)

 - get your own personal FilteredTree and apply on it the filters you
   want without interfering with the main view. (This is how the closed
   tasks pane is built currently)

"""


#=== IMPORT ====================================================================
import os
from xdg.BaseDirectory import xdg_data_home, xdg_config_home
from configobj         import ConfigObj
from GTG.tools.testingmode import TestingMode

import GTG
from GTG.tools.logger import Log
from GTG.tools.borg   import Borg



class CoreConfig(Borg):
    

    #The projects and tasks are of course DATA !
    #We then use XDG_DATA for them
    #Don't forget the "/" at the end.
    DATA_FILE = "projects.xml"
    CONF_FILE = "gtg.conf"
    TASK_CONF_FILE = "tasks.conf"
    conf_dict = None
    #DBUS
    BUSNAME = "org.gnome.GTG"
    BUSINTERFACE = "/org/gnome/GTG"
    #TAGS
    ALLTASKS_TAG = "gtg-tags-all"
    NOTAG_TAG = "gtg-tags-none"
    SEP_TAG = "gtg-tags-sep"

    def __init__(self):
        if  hasattr(self, 'data_dir'):
            #Borg has already been initialized
            return
        if TestingMode().get_testing_mode():
            #we avoid running tests in the user data dir
            self.data_dir = '/tmp/GTG_TESTS/data'
            self.conf_dir = '/tmp/GTG_TESTS/conf'
        else:
            self.data_dir = os.path.join(xdg_data_home,'gtg/')
            self.conf_dir = os.path.join(xdg_config_home,'gtg/')
        if not os.path.exists(self.conf_dir):
            os.makedirs(self.conf_dir)
        if not os.path.exists(self.data_dir):
            os.makedirs(self.data_dir)
        if not os.path.exists(self.conf_dir + self.CONF_FILE):
            f = open(self.conf_dir + self.CONF_FILE, "w")
            f.close()
        if not os.path.exists(self.conf_dir + self.TASK_CONF_FILE):
            f = open(self.conf_dir + self.TASK_CONF_FILE, "w")
            f.close()
        for file in [self.conf_dir + self.CONF_FILE,
                     self.conf_dir + self.TASK_CONF_FILE]:
            if not ((file, os.R_OK) and os.access(file, os.W_OK)):
                raise Exception("File " + file + \
                            " is a configuration file for gtg, but it "
                            "cannot be read or written. Please check it")
        self.conf_dict = ConfigObj(self.conf_dir + self.CONF_FILE)
        self.task_conf_dict = ConfigObj(self.conf_dir + self.TASK_CONF_FILE)
    
    def save(self):
        ''' Saves the configuration of CoreConfig '''
        self.conf_dict.write()
        self.task_conf_dict.write()

    def get_icons_directories(self):
        '''
        Returns the directories containing the icons
        '''
        return [GTG.DATA_DIR, os.path.join(GTG.DATA_DIR, "icons")]

    def get_data_dir(self):
        return self.data_dir

    def set_data_dir(self, path):
        self.data_dir = path

    def get_conf_dir(self):
        return self.conf_dir

    def set_conf_dir(self, path):
        self.conf_dir = path
